#!/usr/bin/env python3

import usb.core
import usb.util
import time
import os

LANG_EN = 0x0409

VENDOR_ID = 0x4242
PRODUCT_ID = 0xe131

MANUFACTURER_INDEX = 1
PRODUCT_INDEX = 2

MANUFACTURER_NAME = 'upb.ro'
PRODUCT_NAME = 'PM Keypad'

# USB_OUT = host to device && Vendor && Device
# USB_IN  = device to host && Vendor && Device
# according to https://www.beyondlogic.org/usbnutshell/usb6.shtml
USB_OUT = 0x40
USB_IN  = 0xC0

VREQ_STATUS            = 0x00
VREQ_GET_BUTTONS       = 0x01
VREQ_SET_STATUS_HID    = 0x02
VREQ_SET_STATUS_VENDOR = 0x03

STATUS_HID    = 0x00
STATUS_VENDOR = 0x01

NUM_BUTTONS = 36

def get_device():
    """
    Returns the device we're looking for (the one
    with our VID/PID and our manufacturer and product
    strings). If it's not found, it returns None
    """
    dev = None

    try:
        dev = usb.core.find(idVendor=VENDOR_ID, idProduct=PRODUCT_ID);
        if dev is None:
            return None

        manufacturer = usb.util.get_string(dev, MANUFACTURER_INDEX, LANG_EN)
        product = usb.util.get_string(dev, PRODUCT_INDEX, LANG_EN)

        if manufacturer != MANUFACTURER_NAME:
            return None

        if product != PRODUCT_NAME:
            return None
    except:
        return None

    return dev


def wait_for_device():
    """
    Tries to get the device every second. It returns
    the device once it finds it.
    """
    dev = None

    # wait until the device connects
    while True:
        print('waiting for the device to connect')
        dev = get_device()
        if dev is not None:
            break
        time.sleep(1)

    print('device connected')
    return dev


def run_script(idx):
    """
    In order to bind a button to a new action, create a new script
    with the following format in the scripts folder. button 0
    is in the top left corner (F1 in the picture), button 1
    is the next one to the right, and so on.
    """
    script_path = '/insert/script/folder/here/script{}.sh'
    script_path = script_path.format(idx)
    script_path += ' & disown'
    os.spawnv(os.P_NOWAIT, '/bin/bash', ['bash', '-c', script_path])


def reap_children():
    """
    reaps processes that were spawned when macros were triggered
    """
    try:
        os.waitid(os.P_ALL, 0, os.WEXITED | os.WNOHANG)
    except:
        pass


def main():
    while True:
        dev = wait_for_device()
        old_status = STATUS_HID

        while True:
            # wait for any children that appeared due to running scripts
            reap_children()
            
            try:
                # get the status of the keyboard (HID/VENDOR)
                ret = dev.ctrl_transfer(USB_IN, VREQ_STATUS, 0, 0, 1);
                status = ret[0]

                if status == STATUS_HID:
                    time.sleep(1)
                    continue

                if status == STATUS_VENDOR and old_status != status:
                    old_buttons = [0] * NUM_BUTTONS

                if status == STATUS_VENDOR:
                    # get the current button status
                    buttons = list(dev.ctrl_transfer(USB_IN,
                                                     VREQ_GET_BUTTONS,
                                                     0, 0, NUM_BUTTONS))
                    # this is used to only trigger macros when a button
                    # is first pressed, you don't want to open 10 calculators
                    # just because you didn't lift your finger quickly enough
                    rising_edge = [e2 - e1 == 1 for e1, e2
                                   in zip(old_buttons, buttons)]
                    old_buttons = buttons

                    for idx, val in enumerate(rising_edge):
                        if val:
                            run_script(idx)

                old_status = status
                time.sleep(0.05)
            except usb.core.USBError:
                break

if __name__ == '__main__':
    main()
