#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include "usbdrv.h"
#include "keydefs.h"


#define USB_REQ_GET_BUTTONS 1
#define NUM_BUTTONS_Y 6
#define NUM_BUTTONS_X 6
#define NUM_BUTTONS (NUM_BUTTONS_Y * NUM_BUTTONS_Y)
#define BUF_LEN NUM_BUTTONS

/* vendor requests */
#define VREQ_STATUS            0x00
#define VREQ_GET_BUTTONS       0x01
#define VREQ_SET_STATUS_HID    0x02
#define VREQ_SET_STATUS_VENDOR 0x03
#define VREQ_TOGGLE_STATUS     0x04

#define STATUS_HID    0x00
#define STATUS_VENDOR 0x01

char return_buf[BUF_LEN];
char buttons_status[BUF_LEN];
int return_buf_index = 0;
uchar status = STATUS_HID;

/* standard boot-enabled keyboard HID descriptor */
PROGMEM const char
usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)(Key Codes)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)(224)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)(231)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs) ; Modifier byte
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x81, 0x03,                    //   INPUT (Cnst,Var,Abs) ; Reserved byte
    0x95, 0x05,                    //   REPORT_COUNT (5)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x05, 0x08,                    //   USAGE_PAGE (LEDs)
    0x19, 0x01,                    //   USAGE_MINIMUM (Num Lock)
    0x29, 0x05,                    //   USAGE_MAXIMUM (Kana)
    0x91, 0x02,                    //   OUTPUT (Data,Var,Abs) ; LED report
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x03,                    //   REPORT_SIZE (3)
    0x91, 0x03,                    //   OUTPUT (Cnst,Var,Abs) ; LED report padding
    0x95, 0x06,                    //   REPORT_COUNT (6)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)(Key Codes)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))(0)
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)(101)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};

typedef struct {
    uint8_t modifier; /* unused in our case */
    uint8_t reserved;
    uint8_t keycode[6];
} keyboard_report_t;

static keyboard_report_t keyboard_report;
volatile static uchar LED_state = 0xff;
static uchar idleRate;

USB_PUBLIC usbMsgLen_t
usbFunctionSetup(uchar data[8])
{
    static uchar usb_ret_buf[BUF_LEN];
    int i;
    usbRequest_t *rq = (void *)data;

    /* class specific messages first */
    if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {
        switch (rq->bRequest) {
        case USBRQ_HID_GET_REPORT:
            usbMsgPtr = (void *)&keyboard_report;
            keyboard_report.modifier = 0;
            keyboard_report.keycode[0] = 0;
        case USBRQ_HID_SET_REPORT: /* wLength == 1 -> LED state (CapsLock etc) */
            return (rq->wLength.word == 1) ? USB_NO_MSG : 0;
        case USBRQ_HID_GET_IDLE: /* required by the standard */
            usbMsgPtr = &idleRate;
            return 1;
        case USBRQ_HID_SET_IDLE: /* required by the standard */
            idleRate = rq->wValue.bytes[1];
            return 0;
        }
    } else if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_VENDOR){
        /* vendor specific messages */
        switch (rq->bRequest) {
        case VREQ_STATUS: /* get current status (HID vs. VENDOR) */
            usbMsgPtr = &status;
            return 1;
        case VREQ_GET_BUTTONS:
            /* returns the buttons pressed since the last request and clears
             * the old button_status */
            for (i = 0; i < BUF_LEN; i++)
                usb_ret_buf[i] = buttons_status[i];
            for (i = 0; i < BUF_LEN; i++)
                buttons_status[i] = 0;
            usbMsgPtr = usb_ret_buf;
            return BUF_LEN;
        case VREQ_SET_STATUS_HID:
            status = STATUS_HID;
            /* reset the button status */
            for (i = 0; i < BUF_LEN; i++)
                buttons_status[i] = 0;
            return 0;
        case VREQ_SET_STATUS_VENDOR:
            status = STATUS_VENDOR;
            /* reset the button status */
            for (i = 0; i < BUF_LEN; i++)
                buttons_status[i] = 0;
            return 0;
        case VREQ_TOGGLE_STATUS:
            status = status == STATUS_HID ? STATUS_VENDOR : STATUS_HID;
            return 0;
        default:
            return 0;
        }
    }

    return 0; /* by default, return 0 */
}

#define RESET_INDEX -1

/* adds a button to return_buf. return_buf is almost like a keycode array, but
 * it can fit all buttons */
static void
add_button(int button)
{
    int i = 0;

    if (button == RESET_INDEX)  {
        /* zero out the buffer and reset the index to the first position */
        return_buf_index = 0;

        for (i = 0; i < BUF_LEN; i++)
            return_buf[i] = 0;

        return;
    }

    return_buf[return_buf_index++] = button;
}

/* applies basic debouncing and save the state of the buttons in button_status
 * (for macros) and return_buf (for HID) */
static void
debounce(void)
{
    uchar i, j, row_start;
    uint16_t presses[NUM_BUTTONS];
    static uint16_t old_presses[NUM_BUTTONS] = {9};

    PORTA = 0x0;
    DDRA = 0x0;

    for (j = 0; j < 6; j++) {
        row_start = j * 6;
        DDRA = (1 << j);
        _delay_us(50);

        for (i = 0; i < 6; i++)
            presses[row_start + i] = PINB & (1 << i) ? 1 : 0;

        DDRA = 0;
        _delay_us(50);
    }
    
    for (i = 0; i < NUM_BUTTONS; i++) {
        /* only add the key if the button is pressed
         * and it was also pressed last time we checked. We can debounce like
         * this because the switch datasheet specifies a 5ms max debounce time
         */
        if ((presses[i] == 0) && (presses[i] == old_presses[i]))
            add_button(i);

        if (presses[i] == old_presses[i] && status == STATUS_VENDOR)
            buttons_status[i] = presses[i] == 0 ? 1 : buttons_status[i];

        old_presses[i] = presses[i];
    }
}

#define NUM_LOCK    1
#define CAPS_LOCK   2
#define SCROLL_LOCK 4

static char keymap[BUF_LEN] = { KEY_F1, KEY_F2, KEY_F3,  KEY_F4, KEY_F5,
    KEY_F6, KEY_F7, KEY_F8, KEY_F9, KEY_F10, KEY_F11, KEY_F12, KEY_1, KEY_2,
    KEY_3, KEY_4, KEY_5, KEY_6, KEY_7, KEY_8, KEY_9, KEY_0, KEY_MINUS,
    KEY_EQUAL, KEY_Q, KEY_W, KEY_E, KEY_R, KEY_T, KEY_Y, KEY_U, KEY_I, KEY_O,
    KEY_P, KEY_SQUARE_BRACKET_OPEN, KEY_SQUARE_BRACKET_CLOSE
};

/* used to set LED state (in my case, it just turns on the on-board LED if
 * caps-lock is pressed) */
usbMsgLen_t usbFunctionWrite(uint8_t *data, uchar len)
{
    if (data[0] == LED_state)
        return 1;
    else
        LED_state = data[0];

    // LED state changed
    if (LED_state & CAPS_LOCK)
        PORTD |=  (1 << PD7); /* LED on  */
    else
        PORTD &= ~(1 << PD7); /* LED off */

    return 1; /* data read, not expecting more */
}

/* builds the HID report from the return_buf array */
static void
build_report()
{
    int i = 0;

    if (return_buf_index == 0) {
        for (i = 0; i < 6; i++)
            keyboard_report.keycode[i] = 0;
        
        return;
    }

    for (i = 0; i < return_buf_index && i < 6; i++)
        keyboard_report.keycode[i] = keymap[(int)return_buf[i]];
}


int __attribute__((noreturn))
main(void)
{
    int i;

    DDRD |= (1 << PD7);
    PORTB = 0x3f;

    /* clear up the report */
    for (i = 0; i < sizeof(keyboard_report); i++)
        ((uchar *)&keyboard_report)[i] = 0;

    for (i = 0; i < BUF_LEN; i++)
        return_buf[i] = 0;

    /* initialize and re-enumerate the device */
    wdt_enable(WDTO_1S);
    usbInit();
    usbDeviceDisconnect();
    for (i = 0; i < 250; i++) {
        wdt_reset();
        _delay_ms(2);
    }
    usbDeviceConnect();
    sei(); /* enable interrupts after re-enumeration */

    // main loop
    while (1) {
        wdt_reset();
        usbPoll();
        debounce(); /* get the button state */
        build_report(); /* build the USB HID report */
        add_button(RESET_INDEX); /* reset the return_buf */

        /* LED_state != 0xff -> the keyboard was surely initialized */
        if (usbInterruptIsReady() &&
            LED_state != 0xff &&
            status == STATUS_HID)
        {
            usbSetInterrupt((void *)&keyboard_report,
                            sizeof(keyboard_report));
        }

        _delay_ms(5);
    }

    /* unreachable */
}

